# optimized statistics functions
# cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False
from cython.view cimport array as cvarray
from libc.stdlib cimport malloc, free
cimport numpy as np


def anova_full_fmaps(
        const np.npy_float64[:,:] y,
        const np.npy_float64[:,:] x,
        const np.npy_float64[:,:] xsinv,
        np.npy_float64[:,:] f_map,
        const np.npy_int64[:,:] effects,
        const np.npy_int8[:, :] e_ms,
):
    """Compute f-maps for a balanced, fully specified ANOVA model
    
    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    x : array (n_cases, n_betas)
        model matrix.
    xsinv : array (n_betas, n_cases)
        xsinv for regression.
    f_map : array (n_fs, n_tests)
        container for output.
    effects : array (n_effects, 2)
        For each effect, indicating the first index in betas and df.
    e_ms : array (n_effects, n_effects)
        Each row represents the expected MS of one effect.
    """
    cdef:
        Py_ssize_t i, i_beta, i_fmap, i_effect, i_effect_ms, i_start, i_stop, case, df
        double v, ss, ms_denom

        Py_ssize_t n_tests = y.shape[1]
        Py_ssize_t n_cases = y.shape[0]
        Py_ssize_t n_betas = x.shape[1]
        Py_ssize_t n_effects = effects.shape[0]
        double *betas = <double *>malloc(sizeof(double) * n_betas)
        double *mss = <double *>malloc(sizeof(double) * n_effects)

    for i in range(n_tests):
        if zero_variance(y, i) == 1:
            for i_fmap in range(f_map.shape[0]):
                f_map[i_fmap, i] = 0
            continue

        _lm_betas(y, i, xsinv, betas)

        # find MS of effects
        for i_effect in range(n_effects):
            i_start = effects[i_effect, 0]
            df = effects[i_effect, 1]
            i_stop = i_start + df
            ss = 0
            for case in range(n_cases):
                v = 0
                for i_beta in range(i_start, i_stop):
                    v += x[case, i_beta] * betas[i_beta]
                ss += v ** 2
            mss[i_effect] = ss / df

        # compute F maps
        i_fmap = 0
        for i_effect in range(n_effects):
            ms_denom = 0
            for i_effect_ms in range(n_effects):
                if e_ms[i_effect, i_effect_ms] > 0:
                    ms_denom += mss[i_effect_ms]

            if ms_denom > 0:
                f_map[i_fmap, i] = mss[i_effect] / ms_denom
                i_fmap += 1

    free(betas)
    free(mss)


def anova_fmaps(const np.npy_float64[:,:] y,
                const np.npy_float64[:,:] x,
                const np.npy_float64[:,:] xsinv,
                np.npy_float64[:,:] f_map,
                const np.npy_int64[:,:] effects,
                int df_res):
    """Compute f-maps for a balanced ANOVA model with residuals

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    x : array (n_cases, n_betas)
        model matrix.
    xsinv : array (n_betas, n_cases)
        xsinv for regression.
    f_map : array (n_fs, n_tests)
        container for output.
    effects : array (n_effects, 2)
        For each effect, indicating the first index in betas and df.
    df_res : int
        Df of the residuals.
    """
    cdef:
        unsigned long i
        unsigned int df, i_beta, i_effect, i_effect_ms, i_effect_beta, i_fmap, case
        double v, SS, MS, MS_res, MS_den

        unsigned long n_tests = y.shape[1]
        unsigned int n_cases = y.shape[0]
        unsigned int n_betas = x.shape[1]
        unsigned int n_effects = effects.shape[0]
        double *betas = <double *>malloc(sizeof(double) * n_betas)
        double [:,:] values = cvarray((n_cases, n_betas), sizeof(double), "d")
        double *predicted_y = <double *>malloc(sizeof(double) * n_cases)

    for i in range(n_tests):
        if zero_variance(y, i) == 1:
            for i_fmap in range(f_map.shape[0]):
                f_map[i_fmap, i] = 0
            continue

        _lm_betas(y, i, xsinv, betas)

        # expand accounted variance
        for case in range(n_cases):
            predicted_y[case] = 0
            for i_beta in range(n_betas):
                v = x[case, i_beta] * betas[i_beta]
                predicted_y[case] += v
                values[case, i_beta] = v

        # residuals
        SS = 0
        for case in range(n_cases):
            SS += (y[case, i] - predicted_y[case]) ** 2
        MS_res = SS / df_res

        # find MS of effects
        for i_effect in range(n_effects):
            i_effect_beta = effects[i_effect, 0]
            df = effects[i_effect, 1]
            SS = 0
            for case in range(n_cases):
                v = 0
                for i_beta in range(i_effect_beta, i_effect_beta + df):
                    v += values[case, i_beta]
                SS += v ** 2
            MS = SS / df
            f_map[i_effect, i] = MS / MS_res

    free(betas)


def sum_square(
        const np.npy_float64[:,:] y,
        np.npy_float64[:] out,
):
    """Compute the Sum Square of the data

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    out : array (n_tests,)
        container for output.
    """
    cdef:
        unsigned long i
        unsigned int case
        double ss

        unsigned long n_tests = y.shape[1]
        unsigned int n_cases = y.shape[0]

    for i in range(n_tests):
        ss = 0
        for case in range(n_cases):
            ss += y[case, i] ** 2

        out[i] = ss


def ss(
        const np.npy_float64[:,:] y,
        np.npy_float64[:] out,
):
    """Compute sum squares in the data (after subtracting the intercept)

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    out : array (n_tests,)
        container for output.
    """
    cdef:
        unsigned long i
        unsigned int case
        double mean, ss_

        unsigned long n_tests = y.shape[1]
        unsigned int n_cases = y.shape[0]

    for i in range(n_tests):
        # find mean
        mean = 0
        for case in range(n_cases):
            mean += y[case, i]
        mean /= n_cases

        # find SS of residuals
        ss_ = 0
        for case in range(n_cases):
            ss_ += (y[case, i] - mean) ** 2

        out[i] = ss_


cdef int zero_variance(
        const np.npy_float64[:,:] y,
        unsigned long i,
) noexcept:
    """Check whether a column of y has zero variance"""
    cdef:
        unsigned int case
        double ref_value = y[0, i]

    for case in range(1, y.shape[0]):
        if y[case, i] != ref_value:
            return 0
    return 1


cdef void _lm_betas(
        const np.npy_float64[:,:] y,
        unsigned long i,
        const np.npy_float64[:,:] xsinv,
        double *betas,
) noexcept:
    """Fit a linear model

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    i : int
        Index of the test for which to calculate betas.
    xsinv : array (n_betas, n_cases)
        xsinv for x.
    betas : array (n_betas,)
        Output container.
    """
    cdef:
        unsigned int i_beta, case
        double beta

        unsigned int n_cases = y.shape[0]
        unsigned int df_x = xsinv.shape[0]

    # betas = xsinv * y
    for i_beta in range(df_x):
        beta = 0
        for case in range(n_cases):
            beta += xsinv[i_beta, case] * y[case, i]
        betas[i_beta] = beta


cdef double _lm_res_ss(
        const np.npy_float64[:,:] y,
        int i,
        const np.npy_float64[:,:] x,
        int df_x,
        double *betas,
) noexcept:
    """Residual sum squares

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    i : int
        Index of the test for which to calculate betas.
    x : array (n_cases, df_model)
        Model matrix.
    betas : array (n_betas,)
        Fitted regression coefficients.
    """
    cdef:
        Py_ssize_t case, i_beta
        double predicted_y

        double ss = 0
        Py_ssize_t n_cases = y.shape[0]

    for case in range(n_cases):
        predicted_y = 0
        for i_beta in range(df_x):
            predicted_y += x[case, i_beta] * betas[i_beta]
        ss += (y[case, i] - predicted_y) ** 2

    return ss


def lm_res(
        const np.npy_float64[:,:] y,
        const np.npy_float64[:,:] x,
        const np.npy_float64[:,:] xsinv,
        np.npy_float64[:,:] res,
):
    """Fit a linear model and compute the residuals

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    x : array (n_cases, n_betas)
        Model matrix for the model.
    xsinv : array (n_betas, n_cases)
        xsinv for x.
    res : array (n_cases, n_tests)
        Container for output.
    """
    cdef:
        Py_ssize_t i, i_beta, case
        double predicted_y, SS_res
        Py_ssize_t n_tests = y.shape[1]
        Py_ssize_t n_cases = y.shape[0]
        Py_ssize_t df_x = xsinv.shape[0]
        double *betas = <double *>malloc(sizeof(double) * df_x)

    for i in range(n_tests):
        _lm_betas(y, i, xsinv, betas)

        # predict y and find residuals
        for case in range(n_cases):
            predicted_y = 0
            for i_beta in range(df_x):
                predicted_y += x[case, i_beta] * betas[i_beta]
            res[case,i] = y[case, i] - predicted_y

    free(betas)


def lm_res_ss(
        const np.npy_float64[:,:] y,
        const np.npy_float64[:,:] x,
        const np.npy_float64[:,:] xsinv,
        np.npy_float64[:] ss,
):
    """Fit a linear model and compute the residual sum squares

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    x : array (n_cases, n_betas)
        Model matrix for the model.
    xsinv : array (n_betas, n_cases)
        xsinv for x.
    ss : array (n_tests,)
        Container for output.
    """
    cdef:
        unsigned long i
        unsigned long n_tests = y.shape[1]
        unsigned int n_cases = y.shape[0]
        unsigned int df_x = xsinv.shape[0]
        double *betas = <double *>malloc(sizeof(double) * df_x)

    for i in range(n_tests):
        _lm_betas(y, i, xsinv, betas)
        ss[i] = _lm_res_ss(y, i, x, df_x, betas)

    free(betas)


def t_1samp(
        const np.npy_float64[:,:] y,
        np.npy_float64[:] out,
):
    """T-values for 1-sample t-test

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    out : array (n_tests,)
        Container for output.
    """
    cdef unsigned long i, case
    cdef double mean, denom
    cdef unsigned long n_tests = y.shape[1]
    cdef unsigned long n_cases = y.shape[0]
    cdef double div = (n_cases - 1) * n_cases

    for i in range(n_tests):
        # mean
        mean = 0
        for case in range(n_cases):
            mean += y[case, i]
        mean /= n_cases

        # variance
        denom = 0
        for case in range(n_cases):
            denom += (y[case, i] - mean) ** 2

        denom /= div
        denom **= 0.5
        if denom > 0:
            out[i] = mean / denom
        else:
            out[i] = 0


def t_1samp_perm(
        const np.npy_float64[:,:] y,
        np.npy_float64[:] out,
        const np.npy_int8[:] sign,
):
    """T-values for 1-sample t-test

    Parameters
    ----------
    y : array (n_cases, n_tests)
        Dependent Measurement.
    out : array (n_tests,)
        Container for output.
    sign : array (n_cases,)
        The randomly asigned sign for each case.
    """
    cdef:
        unsigned long i, case
        double mean, denom
        unsigned long n_tests = y.shape[1]
        unsigned int n_cases = y.shape[0]
        double div = (n_cases - 1) * n_cases
        double *case_buffer = <double *>malloc(sizeof(double) * n_cases)

    for i in range(n_tests):
        for case in range(n_cases):
            case_buffer[case] = y[case, i] * sign[case]

        # mean
        mean = 0
        for case in range(n_cases):
            mean += case_buffer[case]
        mean /= n_cases

        # variance
        denom = 0
        for case in range(n_cases):
            denom += (case_buffer[case] - mean) ** 2

        denom /= div
        denom **= 0.5
        if denom > 0:
            out[i] = mean / denom
        else:
            out[i] = 0


def t_ind(
        const np.npy_float64[:,:] y,
        np.npy_float64[:] out,
        const np.npy_int8[:] group,
):
    "Indpendent-samples t-test, assuming equal variance"
    cdef:
        Py_ssize_t i, case
        double mean0, mean1, var
        Py_ssize_t n_tests = y.shape[1]
        Py_ssize_t n_cases = y.shape[0]
        Py_ssize_t df = n_cases - 2
        Py_ssize_t n0
        Py_ssize_t n1 = 0

    if group.shape[0] != n_cases:
        raise ValueError("length of group does not match n_cases in y")

    for case in range(n_cases):
        if group[case]:
            n1 += 1
    n0 = n_cases - n1

    cdef double var_mult = (1. / n0 + 1. / n1) / df

    for i in range(n_tests):
        mean0 = 0.
        mean1 = 0.
        var = 0.

        # means
        for case in range(n_cases):
            if group[case]:
                mean1 += y[case, i]
            else:
                mean0 += y[case, i]
        mean0 /= n0
        mean1 /= n1

        # variance
        for case in range(n_cases):
            if group[case]:
                var += (y[case, i] - mean1) ** 2
            else:
                var += (y[case, i] - mean0) ** 2
        if var == 0:
            out[i] = 0
            continue
        out[i] = (mean1 - mean0) / (var * var_mult) ** 0.5


def has_zero_variance(const np.npy_float64[:,:] y):
    "True if any data-columns have zero variance"
    cdef:
        double value
        Py_ssize_t case, i

    for i in range(y.shape[1]):
        value = y[0, i]
        for case in range(1, y.shape[0]):
            if y[case, i] != value:
                break
        else:
            return True
    return False



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np

from eelbrain import datasets, plot, testnd, cwt_morlet
from eelbrain.plot._figure import Figure
from eelbrain.plot._utsnd import _ax_bfly_epoch
from eelbrain.testing import requires_mne_sample_data
from eelbrain.testing import hide_plots


@hide_plots
def test_plot_butterfly():
    "Test plot.Butterfly"
    ds = datasets.get_uts(utsnd=True)
    p = plot.Butterfly('utsnd', data=ds)
    p.close()
    p = plot.Butterfly('utsnd', 'A%B', data=ds)
    p.close()

    # other y-dim
    stc = datasets.get_mne_stc(True, subject='fsaverage')
    p = plot.Butterfly(stc)
    p.close()

    # _ax_bfly_epoch (used in GUI, not part of a figure)
    fig = Figure(1)
    ax = _ax_bfly_epoch(fig.axes[0], ds[0, 'utsnd'])
    fig.show()
    ax.set_data(ds[1, 'utsnd'])
    fig.draw()

    # source space
    stc = datasets.get_mne_stc(True)
    p = plot.Butterfly(stc, '.source.hemi')
    p.close()


@hide_plots
def test_plot_array():
    "Test plot.Array"
    ds = datasets.get_uts(utsnd=True)
    p = plot.Array('utsnd', data=ds)
    p.close()
    p = plot.Array('utsnd', 'A%B', data=ds)
    assert p._layout.nax == 4
    assert (p._layout.rows, p._layout.columns) == (2, 2)
    p.close()
    p = plot.Array('utsnd', 'A', sub='B=="b1"', data=ds)
    assert p._layout.nax == 2
    p.close()

    # Scalar dimension
    sgram = abs(cwt_morlet(ds[0, 'uts'], np.arange(3, 7, 0.1) ** 2))
    p = plot.Array(sgram)
    labels = [l.get_text() for l in p.figure.axes[0].get_yticklabels()]
    assert labels == ['9', '12', '15', '19', '23', '27', '32', '38', '44']
    p.close()

    sgram = abs(cwt_morlet(ds[0, 'uts'], np.arange(3, 3.5, 0.01) ** 2))
    p = plot.Array(sgram)
    labels = [l.get_text() for l in p.figure.axes[0].get_yticklabels()]
    assert labels == ['9.0', '9.4', '9.7', '10.0', '10.4', '10.8', '11.1', '11.4', '11.8', '12.2']
    p.close()

    # From test results
    res1 = testnd.TTestIndependent('utsnd', 'A', 'a1', 'a0', data=ds, tstart=0, samples=10)
    res2 = testnd.TTestIndependent('utsnd', 'B', 'b1', 'b0', data=ds, tstart=0, samples=10)
    plot.TopoArray([res1, res2])

    # Categorial dimension
    ds = datasets._get_continuous()
    p = plot.Array(ds['x2'], interpolation='none')
    assert len(p.figure.axes[0].get_yticks()) == 2


@requires_mne_sample_data
@hide_plots
def test_plot_mne_evoked():
    "Test plotting evoked from the mne sample dataset"
    evoked = datasets.get_mne_evoked()
    p = plot.Array(evoked)
    p.close()


@requires_mne_sample_data
@hide_plots
def test_plot_mne_epochs():
    "Test plotting epochs from the mne sample dataset"
    epochs = datasets.get_mne_epochs()

    # grand average
    p = plot.Array(epochs)
    p.close()

    # with model
    p = plot.Array(epochs, np.arange(2).repeat(8))
    p.close()


@hide_plots
def test_plot_results():
    "Test plotting test results"
    ds = datasets.get_uts(True)

    # ANOVA
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, samples=0, pmin=0.05)
    p = plot.Array(res)
    p.close()
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, samples=2, pmin=0.05)
    p = plot.Array(res)
    p.close()

    # Correlation
    res = testnd.Correlation('utsnd', 'Y', 'rm', data=ds)
    p = plot.Array(res)
    p.close()
    res = testnd.Correlation('utsnd', 'Y', 'rm', data=ds, samples=10, pmin=0.05)
    p = plot.Array(res)
    p.close()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from typing import Tuple, Union

import numpy as np

from .. import fmtxt, plot
from .._data_obj import cellname
from .._stats.testnd import TTestIndependent, TTestRelated
from .._text import ms
from ..fmtxt import Section
from ._source import cluster_timecourse


def sensor_results(res, ds, color):
    report = Section("Results")
    if res._kind == 'cluster':
        p = plot.Topomap(res, show=False)
        report.add_figure("Significant clusters.", p)
        p.close()

        report.add_figure("All clusters.", res.clusters)
    else:
        raise NotImplementedError("Result kind %r" % res._kind)
    return report


def sensor_time_frequency_results(
        res: Union[TTestIndependent, TTestRelated],
        p: float = 0.05,
        xlim: Tuple[float, float] = None,
        vmax: float = None,
) -> fmtxt.FMText:
    if res.p.min() > p:
        return fmtxt.asfmtext(res)
    clusters = res.find_clusters(p, True)
    table_ = fmtxt.Table('lll')
    table_.cells('ID', 'Sensors', 'Difference in cluster')
    for cid, cluster_map, frequency_min, frequency_max, t_start, t_stop in clusters.zip('id', 'cluster', 'frequency_min', 'frequency_max', 'tstart', 'tstop'):
        table_.cell(cid)
        chs = cluster_map.sensor.names[cluster_map.any(('frequency', 'time'))]
        frequency_window = (frequency_min, frequency_max + 0.001)
        topo = res.difference.mean(frequency=frequency_window, time=(t_start, t_stop))
        frequency_desc = f'{frequency_min:.1f} - {frequency_max:.1f} Hz'
        plot_ = plot.Topomap(topo, mark=chs, mcolor='yellow', clip='circle', title=frequency_desc, vmax=vmax, h=2)
        table_.cell(plot_.image(close=True))
        diff = res.difference.mean(sensor=chs).mask(res.p.min(sensor=chs) > p, missing=True)
        ys = [res.c1_mean.mean(sensor=chs), res.c0_mean.mean(sensor=chs), diff]
        plot_ = plot.Array(ys, columns=3, xlim=xlim, vmax=vmax, h=2)
        table_.cell(plot_.image(close=True))
    return table_


def sensor_time_results(res, ds, colors, include=1):
    y = ds.eval(res.y)
    if res._kind in ('raw', 'tfce'):
        report = Section("Results")
        section = report.add_section("P<=.05")
        sensor_bin_table(section, res, 0.05)
        clusters = res.find_clusters(0.05, maps=True)
        clusters.sort('tstart')
        for cluster in clusters.itercases():
            sensor_time_cluster(section, cluster, y, res._plot_model(), ds,
                                colors, res.match)

        # trend section
        section = report.add_section("Trend: p<=.1")
        sensor_bin_table(section, res, 0.1)

        # not quite there section
        section = report.add_section("Anything: P<=.2")
        sensor_bin_table(section, res, 0.2)
    elif res._kind == 'cluster':
        report = Section("Clusters")
        sensor_bin_table(report, res)
        clusters = res.find_clusters(include, maps=True)
        clusters.sort('tstart')
        for cluster in clusters.itercases():
            sensor_time_cluster(report, cluster, y, res._plot_model(), ds,
                                colors, res.match)
    else:
        raise NotImplementedError("Result kind %r" % res._kind)
    return report


def sensor_bin_table(section, res, pmin=1):
    if pmin == 1:
        caption = "All clusters"
    else:
        caption = "p <= %.s" % pmin

    for effect, cdist in res._iter_cdists():
        ndvar = cdist.masked_parameter_map(pmin)
        if not ndvar.any():
            if effect:
                text = '%s: nothing\n' % effect
            else:
                text = 'Nothing\n'
            section.add_paragraph(text)
            continue
        elif effect:
            caption_ = "%s: %s" % (effect, caption)
        else:
            caption_ = caption
        p = plot.TopomapBins(ndvar, show=False)
        section.add_image_figure(p, caption_)


def sensor_time_cluster(section, cluster, y, model, ds, colors, match='subject'):
    # cluster properties
    tstart_ms = ms(cluster['tstart'])
    tstop_ms = ms(cluster['tstop'])

    # section/title
    p_str = fmtxt.peq(cluster['p'], stars=True)
    title = f"{tstart_ms}-{tstop_ms} {p_str}"
    if 'effect' in cluster:
        title += f" {cluster['effect']}"
    section = section.add_section(title)

    # description
    paragraph = section.add_paragraph("Id %i" % cluster['id'])
    if 'v' in cluster:
        paragraph.append(", v=%s" % cluster['v'])

    # add cluster image to report
    topo = y.summary(time=(cluster['tstart'], cluster['tstop']))
    cluster_topo = cluster['cluster'].any('time')
    cluster_topo.info['contours'] = {0.5: (1, 1, 0)}
    if model:
        x = ds.eval(model)
        topos = [[topo[x == cell].summary('case', name=cellname(cell)),
                  cluster_topo] for cell in x.cells]
    else:
        topos = [[topo, cluster_topo]]
    p = plot.Topomap(topos, axh=3, rows=1, show=False)
    p.mark_sensors(np.flatnonzero(cluster_topo.x), c='y', marker='o')

    caption_ = ["Cluster"]
    if 'effect' in cluster:
        caption_.extend(('effect of', cluster['effect']))
    caption_.append("%i - %i ms." % (tstart_ms, tstop_ms))
    caption = ' '.join(caption_)
    section.add_image_figure(p, caption)
    p.close()

    cluster_timecourse(section, cluster, y, 'sensor', model, ds, colors, match)



from contextlib import ContextDecorator

from appnope import NSActivityUserInitiated, beginActivityWithOptions, endActivity


class ActivityContext(ContextDecorator):
    """Context disabling idle sleep and App Nap on macOS"""
    def __init__(self, options, message):
        self.n_processes = 0
        self.options = options
        self.message = message

    def __enter__(self):
        if self.n_processes == 0:
            self._activity = beginActivityWithOptions(self.options, self.message)
        self.n_processes += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.n_processes -= 1
        if self.n_processes == 0:
            endActivity(self._activity)


user_activity = ActivityContext(NSActivityUserInitiated, 'Eelbrain user activity')



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"Utilities for testing"
from contextlib import ContextDecorator, contextmanager
from functools import reduce
from importlib.util import spec_from_file_location, module_from_spec
import os
from operator import mul
from pathlib import Path
import shutil
import tempfile

import mne
import numpy as np
from numpy.testing import assert_array_equal
import pytest

import eelbrain._wxgui
from .._config import CONFIG
from .._data_obj import Dataset, NDVar, Var, Factor, isdatalist, isuv


class TempDir(str):
    "After MNE-Python mne.utils"
    def __new__(cls):
        return str.__new__(cls, tempfile.mkdtemp())

    def __del__(self):
        shutil.rmtree(self, ignore_errors=True)


def assert_dataset_equal(ds1, ds2, decimal=None):
    """
    Raise an assertion if two Datasets are not equal up to desired precision.

    Parameters
    ----------
    ds1, ds2 : Dataset
        Datasets to compare.
    decimal : None | int
        Desired precision (default is exact match).
    """
    assert ds1.keys() == ds2.keys()
    for k in ds1.keys():
        assert_dataobj_equal(ds1[k], ds2[k], decimal=decimal)
    assert ds1.info.keys() == ds2.info.keys()


def assert_dataobj_equal(d1, d2, decimal=None, name=True):
    """Assert that two data-objects are equal up to desired precision.

    Parameters
    ----------
    d1, d2 : data-objects
        Data-objects to compare.
    decimal : None | int
        Desired precision (default is exact match).
    name : bool
        Assert that ``d1.name == d2.name``.
    """
    assert type(d1) == type(d2)
    if name:
        assert d1.name == d2.name
    assert len(d1) == len(d2)
    if isuv(d1, interaction=True):
        if isinstance(d1, Var):
            is_equal = np.isclose(d1.x, d2.x, equal_nan=True, rtol=0, atol=10**-decimal if decimal else 0)
        else:
            is_equal = d1 == d2
        if not np.all(is_equal):
            ds = Dataset()
            ds['value'] = d1.copy()
            ds['target'] = d2.copy()
            ds['unequal'] = Factor(is_equal, labels={True: '', False: 'x'})
            if isinstance(d1, Var):
                ds['difference'] = d1 - d2
            raise AssertionError(f'Two {d1.__class__.__name__}s named {d1.name!r} have unequal values:\n\n{ds}')
        if isinstance(d1, Factor):
            if d1.random != d2.random:
                raise AssertionError(f"{d1.name}: d1.random={d1.random}, d2.random={d2.random}")
    elif isinstance(d1, NDVar):
        assert d1.dims == d2.dims
        if decimal:
            is_different = np.max(np.abs(d1.x - d2.x)) >= 10**-decimal
        else:
            is_different = np.any(d1.x != d2.x)

        if is_different:
            n = reduce(mul, d1.x.shape)
            difference = np.abs(d1.x - d2.x)
            if decimal:
                different = difference >= 10**-decimal
            else:
                different = d1.x != d2.x
            n_different = different.sum()
            mean_diff = difference[different].sum() / n_different
            raise AssertionError(f"NDVars named {d1.name!r} have unequal values. Difference in {n_different} of {n} values, average difference={mean_diff}.")
    elif isdatalist(d1):
        assert all(item1 == item2 for item1, item2 in zip(d1, d2))
    elif isinstance(d1, Dataset):
        assert_dataset_equal(d1, d2, decimal)
    else:
        raise TypeError(f"{d1.__class__.__name__} is not a data-object type: {d1!r}")


def assert_source_space_equal(src1, src2):
    """Assert that two SourceSpace objects are identical

    Parameters
    ----------
    src1, src2 : SourceSpace objects
        SourceSpace objects to compare.
    """
    assert_array_equal(src1.vertices[0], src2.vertices[0])
    assert_array_equal(src1.vertices[1], src2.vertices[1])
    assert src1.subject == src2.subject
    assert src1.src == src2.src
    assert src1.subjects_dir == src2.subjects_dir


class GUITestContext(ContextDecorator):
    modules = (
        eelbrain._wxgui.select_epochs,
        eelbrain._wxgui.select_components,
        eelbrain._wxgui.history,
        eelbrain._wxgui.load_stcs,
    )

    def __init__(self):
        self._i = 0

    def __enter__(self):
        self._i += 1
        if self._i == 1:
            for mod in self.modules:
                mod.TEST_MODE = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._i -= 1
        if self._i == 0:
            for mod in self.modules:
                mod.TEST_MODE = False


gui_test_context = GUITestContext()


def gui_test(function):
    return gui_test_context(requires_framework_build(function))


class ConfigContext(ContextDecorator):

    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.old = None

    def __enter__(self):
        self.old = CONFIG[self.key]
        CONFIG[self.key] = self.value

    def __exit__(self, exc_type, exc_val, exc_tb):
        CONFIG[self.key] = self.old


hide_plots_context = ConfigContext('show', False)


def hide_plots(function):
    return hide_plots_context(requires_framework_build(function))


@contextmanager
def working_directory(wd):
    "Context for temporarily changing the working directory"
    cwd = os.getcwd()
    os.chdir(wd)
    try:
        yield
    finally:
        os.chdir(cwd)


def import_attr(path, attr):
    spec = spec_from_file_location('module', path)
    mod = module_from_spec(spec)
    spec.loader.exec_module(mod)
    return getattr(mod, attr)


def requires_framework_build(function):
    return pytest.mark.framework_build(function)


def requires_mne_sample_data(function):
    if mne.datasets.sample.data_path(download=False):
        return function
    else:
        return pytest.mark.skip('mne sample data unavailable')(function)


def requires_mne_testing_data(function):
    if mne.datasets.testing.data_path(download=False):
        return function
    else:
        return pytest.mark.skip('mne testing data unavailable')(function)


def requires_pyarrow(function):
    "Sometimes broken under env-dev on Unix"
    try:
        import pyarrow
        return function
    except ImportError:
        return pytest.mark.skip('pyarrow import error')(function)


def requires_r_ez(function):
    from .._utils.r_bridge import r, r_warning_filter

    with r_warning_filter:
        success = r('require(ez)')[0]

    if success:
        return function
    else:
        return pytest.mark.skip('r-ez unavailable')(function)


def skip_on_windows(function):
    if os.name == 'nt':
        return pytest.mark.skip('Test disabled on Windows')(function)
    else:
        return function


def file_path(name):
    "Path to test data file in the test_data directory"
    root = Path(__file__).parents[2]
    if name == 'fox-prestige':
        path = root / 'examples' / 'statistics' / 'Fox_Prestige_data.txt'
    else:
        path = root / 'test_data' / name

    if path.exists():
        return path
    else:
        raise IOError("Testing file does not exist. Test can only be executed from source repository.")


def path(string):
    "OS-independent path check"
    return str(Path(string))



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Migrate cache structures from older versions"""
from itertools import chain
import os
from pathlib import Path


def squeeze_spaces_in_paths(root: Path):
    "Remove double spaces and trailing spaces in path structure"
    for dirpath, dirnames, filenames in os.walk(root, topdown=False):
        for old in chain(filenames, dirnames):
            new = ' '.join(old.split())
            if new != old:
                src = os.path.join(dirpath, old)
                dst = os.path.join(dirpath, new)
                if os.path.exists(dst):
                    src_rel = os.path.relpath(src, root)
                    raise RuntimeError(f"Trying to rename {src_rel} to {new}, which already exists")
                os.rename(src, dst)



from itertools import chain
import pytest

from eelbrain import datasets, plot
from eelbrain.plot import _base
from eelbrain.plot._base import Layout, ImLayout
from eelbrain.testing import skip_on_windows
from eelbrain.testing import hide_plots


def assert_layout_consistent(layout):
    if layout.margins is None:
        return
    assert all(v >= 0 for v in layout.margins.values())
    assert layout.w == layout.margins['left'] + layout.columns * layout.axw + (layout.columns - 1) * layout.margins['wspace'] + layout.margins['right']
    assert layout.h == layout.margins['bottom'] + layout.rows * layout.axh + (layout.rows - 1) * layout.margins['hspace'] + layout.margins['top']


def assert_layout_ok(*args, **kwargs):
    layout = Layout(*args, **kwargs)
    assert layout.rows * layout.columns >= layout.nax
    if layout.h_fixed:
        assert layout.rows * layout.columns < layout.nax + layout.rows
    else:
        assert layout.rows * layout.columns < layout.nax + layout.columns
    assert_layout_consistent(layout)


def test_layout():
    "Test the Layout class"
    for nax in range(1, 100):
        assert_layout_ok(nax, 1.5, 2, True, w=5)
        assert_layout_ok(nax, 1.5, 2, True, h=5)
        assert_layout_ok(nax, 1.5, 2, True, axw=5)
        assert_layout_ok(nax, 1.5, 2, True, axh=5)
        assert_layout_ok(nax, 1.5, 2, True, axw=5, w=20)
        assert_layout_ok(nax, 1.5, 2, True, axw=5, h=20)
        assert_layout_ok(nax, 1.5, 2, True, axh=5, w=20)
        assert_layout_ok(nax, 1.5, 2, True, axh=5, h=20)

    # single axes larger than figure
    with pytest.raises(ValueError):
        Layout(2, 1.5, 2, True, h=5, axh=6)
    with pytest.raises(ValueError):
        Layout(2, 1.5, 2, True, w=5, axw=6)

    # left margin & axw
    margins = dict(left=1, top=2, bottom=1, wspace=1, hspace=2)
    layout = Layout(2, 2, 2, margins=margins, axw=5, w=10, columns=1)
    assert_layout_consistent(layout)
    assert layout.w == 10
    assert layout.margins == dict(right=10 - 1 - 5, **margins)
    assert layout.axh == 2.5
    assert layout.h == 5 + 2 * layout.axh
    assert layout.tight is False

    layout = Layout(2, 2, 2, margins=True, w=10, columns=1)
    assert_layout_consistent(layout)
    assert layout.tight is False

    layout = Layout(1, 1.5, 3, h=2, w=15, margins={'left': 1, 'right': 0.1})
    assert_layout_consistent(layout)


def test_im_layout():
    "Test the ImLayout class"
    l = ImLayout(1, 1, 5, None, {}, w=3)
    assert l.w == 3
    assert l.h == 3
    assert l.axw == l.axh == 3
    l = ImLayout(2, 1, 5, None, {}, w=3, columns=2)
    assert l.w == 3
    assert l.h == 1.5
    assert l.axw == l.axh == 1.5
    l = ImLayout(1, 1, 5, None, {}, axw=3)
    assert l.w == 3
    assert l.h == 3
    assert l.axw == l.axh == 3
    l = ImLayout(2, 1, 5, None, {}, axw=3, columns=2)
    assert l.w == 6
    assert l.h == 3
    assert l.axw == l.axh == 3
    # 2 x 2 layout
    margins = {'left': 0.4, 'right': 0.1, 'top': 0.1, 'bottom': 0.4}
    l = ImLayout(4, 4/3, 2, margins, {'bottom': 0.5}, rows=2, columns=2, w=4, h=4)
    assert l.w == 4
    assert l.axw == 1.75


@hide_plots
def test_time_slicer():
    "Test linked time axes"
    ds = datasets.get_uts(True)

    p1 = plot.Butterfly(ds['utsnd'])
    p2 = plot.Array('utsnd', 'A', data=ds)
    p1.link_time_axis(p2)

    p1._set_time(.1, True)
    assert p2._current_time == .1
    assert p2._time_fixed == True
    p2._set_time(.2)
    assert p1._current_time == .2
    assert p1._time_fixed == False

    p1 = plot.TopoButterfly(ds['utsnd'])
    p2 = plot.Array('utsnd', 'A', data=ds)
    p2.link_time_axis(p1)

    p1._set_time(.1, True)
    assert p2._current_time == .1
    assert p2._time_fixed == True

    # merge another
    p3 = plot.TopoButterfly(ds[0, 'utsnd'])
    p3.link_time_axis(p2)

    p2._set_time(.2)
    assert p1._current_time == .2
    assert p1._time_fixed == False


def test_vlims():
    "Test vlim determination"
    ds = datasets.get_uts()
    epochs = [[ds[i: i+5, 'uts'].mean('case')] for i in range(0, 10, 5)]
    meas = ds['uts'].info.get('meas')

    # without cmap
    lims = _base.find_fig_vlims(epochs)
    assert lims[meas] == (-1, 3)
    lims = _base.find_fig_vlims(epochs, 1)
    assert lims[meas] == (-1, 1)
    lims = _base.find_fig_vlims(epochs, .1)
    assert lims[meas] == (-.1, .1)
    lims = _base.find_fig_vlims(epochs, 1, -2)
    assert lims[meas] == (-2, 1)

    # positive data
    epochs = [[e * e.sign()] for e in chain(*epochs)]
    lims = _base.find_fig_vlims(epochs)
    assert lims[meas][0] == 0
    lims = _base.find_fig_vlims(epochs, 1)
    assert lims[meas] == (0, 1)
    lims = _base.find_fig_vlims(epochs, 1, -1)
    assert lims[meas] == (-1, 1)

    # symmetric
    cmaps = _base.find_fig_cmaps(epochs)
    assert cmaps == {meas: 'xpolar'}
    lims = _base.find_fig_vlims(epochs, cmaps=cmaps)
    assert lims[meas][0] == -lims[meas][1]
    lims = _base.find_fig_vlims(epochs, 1, cmaps=cmaps)
    assert lims[meas] == (-1, 1)
    lims = _base.find_fig_vlims(epochs, 1, 0, cmaps=cmaps)
    assert lims[meas] == (-1, 1)

    # zero-based
    cmaps[meas] = 'sig'
    lims = _base.find_fig_vlims(epochs, cmaps=cmaps)
    assert lims[meas][0] == 0
    lims = _base.find_fig_vlims(epochs, 1, cmaps=cmaps)
    assert lims[meas] == (0, 1)
    lims = _base.find_fig_vlims(epochs, 1, -1, cmaps=cmaps)
    assert lims[meas] == (0, 1)


@skip_on_windows  # window resizes to screen size
@hide_plots
def test_eelfigure():
    ds = datasets.get_uts()

    p = plot.UTSStat('uts', 'A', data=ds, h=2, w=50)
    assert tuple(p.figure.get_size_inches()) == (50, 2)
